﻿Imports System.Windows.Media.Media3D
Imports System.Windows.Media
Imports System.Windows
Imports System.Windows.Shapes
Imports System.Windows.Media.Imaging

Namespace ParticleSample
    Public Class ParticleSystemManager
        Private particleSystems As Dictionary(Of Color, ParticleSystem)

        Public Sub New()
            Me.particleSystems = New Dictionary(Of Color, ParticleSystem)()
        End Sub

        Public Sub Update(elapsed As Single)
            For Each ps As ParticleSystem In Me.particleSystems.Values
                ps.Update(elapsed)
            Next
        End Sub

        Public Sub SpawnParticle(position As Point3D, speed As Double, color As Color, size As Double, life As Double)
            Try
                Dim ps As ParticleSystem = Me.particleSystems(color)
                ps.SpawnParticle(position, speed, size, life)
            Catch
            End Try
        End Sub

        Public Function CreateParticleSystem(maxCount As Integer, color As Color) As Model3D
            Dim ps As New ParticleSystem(maxCount, color)
            Me.particleSystems.Add(color, ps)
            Return ps.ParticleModel
        End Function
    End Class

    Public Class ParticleSystem
        Public particleList As List(Of Particle)
        Public m_particleModel As GeometryModel3D
        Public m_maxParticleCount As Integer
        Public rand As Random

        Public Sub New(maxCount As Integer, color__1 As Color)
            Me.m_maxParticleCount = maxCount

            Me.particleList = New List(Of Particle)()

            Me.m_particleModel = New GeometryModel3D()
            Me.m_particleModel.Geometry = New MeshGeometry3D()

            Dim e As New Ellipse()
            e.Width = 64.0
            e.Height = 64.0
            Dim b As New RadialGradientBrush()
            b.GradientStops.Add(New GradientStop(Color.FromArgb(&HFF, color__1.R, color__1.G, color__1.B), 0.25)) '泡の色
            b.GradientStops.Add(New GradientStop(Color.FromArgb(&H0, color__1.R, color__1.G, color__1.B), 1.0))
            e.Fill = b
            e.Measure(New Size(64, 64))
            e.Arrange(New Rect(0, 0, 64, 64))
            'e.StrokeThickness = 1
            'e.Stroke = New SolidColorBrush(Colors.White) '泡の色
            'e.Measure(New Size(32, 32)) '泡の大きさ
            'e.Arrange(New Rect(0, 0, 32, 32))

            Dim brush As Brush = Nothing

            Dim renderTarget As New RenderTargetBitmap(64, 64, 96, 96, PixelFormats.Pbgra32)
            renderTarget.Render(e)
            renderTarget.Freeze()
            brush = New ImageBrush(renderTarget)
            Dim material As New DiffuseMaterial(brush)

            Me.m_particleModel.Material = material

            Me.rand = New Random(brush.GetHashCode())
        End Sub

        Public Sub Update(elapsed As Double)
            Dim deadList As New List(Of Particle)()
            For Each p As Particle In Me.particleList
                p.Position += p.Velocity * elapsed
                p.Life -= p.Decay * elapsed
                p.Size = p.StartSize * (p.Life / p.StartLife)
                If p.Life <= 0.0 Then
                    deadList.Add(p)
                End If
            Next

            For Each p As Particle In deadList
                Me.particleList.Remove(p)
            Next

            UpdateGeometry()
        End Sub

        Private Sub UpdateGeometry()
            Dim positions As New Point3DCollection()
            Dim indices As New Int32Collection()
            Dim texcoords As New PointCollection()

            For i As Integer = 0 To Me.particleList.Count - 1
                Dim positionIndex As Integer = i * 4
                Dim indexIndex As Integer = i * 6
                Dim p As Particle = Me.particleList(i)

                Dim p1 As New Point3D(p.Position.X, p.Position.Y, p.Position.Z)
                Dim p2 As New Point3D(p.Position.X, p.Position.Y + p.Size, p.Position.Z)
                Dim p3 As New Point3D(p.Position.X + p.Size, p.Position.Y + p.Size, p.Position.Z)
                Dim p4 As New Point3D(p.Position.X + p.Size, p.Position.Y, p.Position.Z)

                positions.Add(p1)
                positions.Add(p2)
                positions.Add(p3)
                positions.Add(p4)

                Dim t1 As New Point(0.0, 0.0)
                Dim t2 As New Point(0.0, 1.0)
                Dim t3 As New Point(1.0, 1.0)
                Dim t4 As New Point(1.0, 0.0)

                texcoords.Add(t1)
                texcoords.Add(t2)
                texcoords.Add(t3)
                texcoords.Add(t4)

                indices.Add(positionIndex)
                indices.Add(positionIndex + 2)
                indices.Add(positionIndex + 1)
                indices.Add(positionIndex)
                indices.Add(positionIndex + 3)
                indices.Add(positionIndex + 2)
            Next

            DirectCast(Me.m_particleModel.Geometry, MeshGeometry3D).Positions = positions
            DirectCast(Me.m_particleModel.Geometry, MeshGeometry3D).TriangleIndices = indices '外すと直線になる
            DirectCast(Me.m_particleModel.Geometry, MeshGeometry3D).TextureCoordinates = texcoords

        End Sub

        Public Sub SpawnParticle(position As Point3D, speed As Double, size As Double, life As Double)
            If Me.particleList.Count > Me.m_maxParticleCount Then
                Return
            End If
            Dim p As New Particle()
            p.Position = position
            p.StartLife = life
            p.Life = life
            p.StartSize = size
            p.Size = size

            Dim x1 As Single = 1.0F - CSng(rand.NextDouble()) * 2.0F
            Dim z As Single = 1.0F - CSng(rand.NextDouble()) * 2.0F

            Dim v As New Vector3D(x1, z, 0.0)
            v.Normalize()
            v *= (CSng(rand.NextDouble()) + 0.25F) * CSng(speed)

            p.Velocity = New Vector3D(v.X, v.Y, v.Z)

            p.Decay = 1.0F
            Me.particleList.Add(p)
        End Sub
        Public pm As New ParticleSystemManager
        Public frameTimer As System.Windows.Threading.DispatcherTimer

        Public spawnPoint As Point3D
        Public elapsed As Double
        Public totalElapsed As Double
        Public lastTick As Integer
        Public currentTick As Integer

        Public frameCount As Integer
        Public frameCountTime As Double
        Public frameRate As Integer

        Public Sub AddPaticle(myElement As Model3DGroup)
            myElement.Children.Add(pm.CreateParticleSystem(1000, Colors.Gray))
            myElement.Children.Add(pm.CreateParticleSystem(1000, Colors.Red))
            myElement.Children.Add(pm.CreateParticleSystem(1000, Colors.Silver))
            myElement.Children.Add(pm.CreateParticleSystem(1000, Colors.Orange))
            myElement.Children.Add(pm.CreateParticleSystem(1000, Colors.Yellow))
        End Sub

        Public Function ParticlePosition(x As Double, y As Double) As Point3D
            Me.currentTick = Environment.TickCount
            Me.elapsed = CDbl(Me.currentTick - Me.lastTick) / 1000.0
            Me.totalElapsed += Me.elapsed
            Me.lastTick = Me.currentTick

            frameCount += 1
            frameCountTime += elapsed
            If frameCountTime >= 1.0 Then
                frameCountTime -= 1.0
                frameRate = frameCount
                frameCount = 0

            End If

            pm.Update(CSng(elapsed))
            pm.SpawnParticle(Me.spawnPoint, 10.0, Colors.Red, rand.NextDouble(), 2.5 * rand.NextDouble())
            pm.SpawnParticle(Me.spawnPoint, 10.0, Colors.Orange, rand.NextDouble(), 2.5 * rand.NextDouble())
            pm.SpawnParticle(Me.spawnPoint, 10.0, Colors.Silver, rand.NextDouble(), 2.5 * rand.NextDouble())
            pm.SpawnParticle(Me.spawnPoint, 10.0, Colors.Gray, rand.NextDouble(), 2.5 * rand.NextDouble())
            pm.SpawnParticle(Me.spawnPoint, 10.0, Colors.Red, rand.NextDouble(), 2.5 * rand.NextDouble())
            pm.SpawnParticle(Me.spawnPoint, 10.0, Colors.Orange, rand.NextDouble(), 2.5 * rand.NextDouble())
            pm.SpawnParticle(Me.spawnPoint, 10.0, Colors.Silver, rand.NextDouble(), 2.5 * rand.NextDouble())
            pm.SpawnParticle(Me.spawnPoint, 10.0, Colors.Gray, rand.NextDouble(), 2.5 * rand.NextDouble())
            pm.SpawnParticle(Me.spawnPoint, 10.0, Colors.Red, rand.NextDouble(), 2.5 * rand.NextDouble())
            pm.SpawnParticle(Me.spawnPoint, 10.0, Colors.Yellow, rand.NextDouble(), 2.5 * rand.NextDouble())
            pm.SpawnParticle(Me.spawnPoint, 10.0, Colors.Silver, rand.NextDouble(), 2.5 * rand.NextDouble())
            pm.SpawnParticle(Me.spawnPoint, 10.0, Colors.Yellow, rand.NextDouble(), 2.5 * rand.NextDouble())
            If x = 0 AndAlso y = 0 Then
                Dim c As Double = Math.Cos(Me.totalElapsed)
                Dim s As Double = Math.Sin(Me.totalElapsed)
                Me.spawnPoint = New Point3D(s * 32.0, c * 32.0, 0.0)
            Else
                Me.spawnPoint = New Point3D((CInt(x)), (CInt(-y)), 0.0)
            End If
            Return spawnPoint
        End Function


        Public ReadOnly Property ParticleModel() As Model3D
            Get
                Return Me.m_particleModel
            End Get
        End Property
    End Class

    Public Class Particle
        Public Position As Point3D
        Public Velocity As Vector3D
        Public StartLife As Double
        Public Life As Double
        Public Decay As Double
        Public StartSize As Double
        Public Size As Double
    End Class

   
End Namespace